@@ -5,25 +5,10 @@
 
 package org.mozilla.gecko.notifications;
 
-import java.io.File;
-import java.io.UnsupportedEncodingException;
-import java.net.URLConnection;
-import java.net.URLDecoder;
-import java.util.List;
-
-import org.mozilla.gecko.AppConstants;
-import org.mozilla.gecko.EventDispatcher;
-import org.mozilla.gecko.GeckoActivityMonitor;
-import org.mozilla.gecko.GeckoAppShell;
-import org.mozilla.gecko.R;
-import org.mozilla.gecko.mozglue.SafeIntent;
-import org.mozilla.gecko.util.BitmapUtils;
-import org.mozilla.gecko.util.BundleEventListener;
-import org.mozilla.gecko.util.EventCallback;
-import org.mozilla.gecko.util.GeckoBundle;
-import org.mozilla.gecko.util.ThreadUtils;
-
+import android.annotation.TargetApi;
 import android.app.Activity;
+import android.app.NotificationChannel;
+import android.app.NotificationManager;
 import android.app.PendingIntent;
 import android.content.ComponentName;
 import android.content.Context;
@@ -36,6 +21,29 @@ import android.support.v4.app.NotificationCompat;
 import android.support.v4.util.SimpleArrayMap;
 import android.util.Log;
 
+import org.mozilla.gecko.AppConstants;
+import org.mozilla.gecko.EventDispatcher;
+import org.mozilla.gecko.GeckoActivityMonitor;
+import org.mozilla.gecko.GeckoAppShell;
+import org.mozilla.gecko.R;
+import org.mozilla.gecko.mozglue.SafeIntent;
+import org.mozilla.gecko.util.BitmapUtils;
+import org.mozilla.gecko.util.BundleEventListener;
+import org.mozilla.gecko.util.EventCallback;
+import org.mozilla.gecko.util.GeckoBundle;
+import org.mozilla.gecko.util.StrictModeContext;
+import org.mozilla.gecko.util.ThreadUtils;
+
+import java.io.File;
+import java.io.UnsupportedEncodingException;
+import java.net.URLConnection;
+import java.net.URLDecoder;
+import java.util.ArrayList;
+import java.util.Arrays;
+import java.util.HashMap;
+import java.util.List;
+import java.util.Map;
+
 public final class NotificationHelper implements BundleEventListener {
     public static final String HELPER_BROADCAST_ACTION = AppConstants.ANDROID_PACKAGE_NAME + ".helperBroadcastAction";
 
@@ -77,6 +85,81 @@ public final class NotificationHelper implements BundleEventListener {
 
     private final Context mContext;
 
+
+    public enum Channel {
+        /**
+         * Default notification channel.
+         */
+        DEFAULT,
+        /**
+         * Mozilla Location Services notification channel.
+         */
+        MLS,
+        /**
+         * File downloads.
+         */
+        DOWNLOAD,
+        /**
+         *  Media notification channel
+         */
+        MEDIA,
+        /**
+         * Synced tabs notification channel
+         */
+        SYNCED_TABS,
+        /**
+         * Leanplum notification channel - use only when <code>AppConstants.MOZ_ANDROID_MMA</code> is true.
+         */
+        LP_DEFAULT,
+        /**
+         * HTML5 web site notifications
+         */
+        SITE_NOTIFICATIONS,
+        /**
+         * Crash handler notification used on Android Q+
+         */
+        CRASH_HANDLER,
+    }
+
+    // Holds the mapping between the Channel enum used by the rest of our codebase and the
+    // channel ID used for communication with the system NotificationManager.
+    // How to determine the initialCapacity: Count all channels (including the Updater, which is
+    // only added further down in initNotificationChannels), multiply by 4/3 for a maximum load
+    // factor of 75 % and round up to the next multiple of two.
+    private final Map<Channel, String> mDefinedNotificationChannels = new HashMap<Channel, String>(15) {{
+        final String DEFAULT_CHANNEL_TAG = "default2-notification-channel";
+        put(Channel.DEFAULT, DEFAULT_CHANNEL_TAG);
+
+        final String MLS_CHANNEL_TAG = "mls-notification-channel";
+        put(Channel.MLS, MLS_CHANNEL_TAG);
+
+        final String DOWNLOAD_NOTIFICATION_TAG = "download-notification-channel";
+        put(Channel.DOWNLOAD, DOWNLOAD_NOTIFICATION_TAG);
+
+        final String MEDIA_CHANNEL_TAG = "media-notification-channel";
+        put(Channel.MEDIA, MEDIA_CHANNEL_TAG);
+
+        if (AppConstants.MOZ_ANDROID_MMA) {
+            final String LP_DEFAULT_CHANNEL_TAG = "lp-default-notification-channel";
+            put(Channel.LP_DEFAULT, LP_DEFAULT_CHANNEL_TAG);
+        }
+
+        final String SYNCED_TABS_CHANNEL_TAG = "synced-tabs-notification-channel";
+        put(Channel.SYNCED_TABS, SYNCED_TABS_CHANNEL_TAG);
+
+        final String SITE_NOTIFICATIONS_CHANNEL_TAG = "site-notifications";
+        put(Channel.SITE_NOTIFICATIONS, SITE_NOTIFICATIONS_CHANNEL_TAG);
+
+        final String CRASH_HANDLER_CHANNEL_TAG = "crash-handler";
+        put(Channel.CRASH_HANDLER, CRASH_HANDLER_CHANNEL_TAG);
+    }};
+
+    // These are channels we no longer require and want to retire from Android's settings UI.
+    private final List<String> mDeprecatedNotificationChannels = new ArrayList<>(Arrays.asList(
+            "default-notification-channel",
+            null
+    ));
+
     // Holds a list of notifications that should be cleared if the Fennec Activity is shut down.
     // Will not include ongoing or persistent notifications that are tied to Gecko's lifecycle.
     private SimpleArrayMap<String, GeckoBundle> mClearableNotifications;
@@ -97,6 +180,11 @@ public final class NotificationHelper implements BundleEventListener {
         EventDispatcher.getInstance().registerUiThreadListener(this,
             "Notification:Show",
             "Notification:Hide");
+
+        if (!AppConstants.Versions.preO) {
+            initNotificationChannels();
+        }
+
         mInitialized = true;
     }
 
@@ -112,6 +200,108 @@ public final class NotificationHelper implements BundleEventListener {
         return sInstance;
     }
 
+    private void initNotificationChannels() {
+        final String UPDATER_CHANNEL_TAG = "updater-notification-channel";
+        mDeprecatedNotificationChannels.add(UPDATER_CHANNEL_TAG);
+
+
+        final NotificationManager notificationManager =
+                (NotificationManager) mContext.getSystemService(Context.NOTIFICATION_SERVICE);
+
+        for (String channelId : mDeprecatedNotificationChannels) {
+            removeChannel(notificationManager, channelId);
+        }
+
+        for (Channel mozChannel : mDefinedNotificationChannels.keySet()) {
+            createChannel(notificationManager, mozChannel);
+        }
+    }
+
+    @SuppressWarnings("fallthrough")
+    @TargetApi(26)
+    private void createChannel(final NotificationManager manager, Channel definedChannel) {
+        NotificationChannel channel =
+                manager.getNotificationChannel(mDefinedNotificationChannels.get(definedChannel));
+
+        if (channel == null) {
+            switch (definedChannel) {
+                case MLS: {
+                    channel = new NotificationChannel(mDefinedNotificationChannels.get(definedChannel),
+                            mContext.getString(R.string.mls_notification_channel),
+                            NotificationManager.IMPORTANCE_LOW);
+                }
+                break;
+
+                case DOWNLOAD: {
+                    channel = new NotificationChannel(mDefinedNotificationChannels.get(definedChannel),
+                            mContext.getString(R.string.download_notification_channel),
+                            NotificationManager.IMPORTANCE_LOW);
+                }
+                break;
+
+                case MEDIA: {
+                    channel = new NotificationChannel(mDefinedNotificationChannels.get(definedChannel),
+                            mContext.getString(R.string.media_notification_channel2),
+                            NotificationManager.IMPORTANCE_LOW);
+                }
+                break;
+
+                case SYNCED_TABS: {
+                    channel = new NotificationChannel(mDefinedNotificationChannels.get(definedChannel),
+                            mContext.getString(R.string.synced_tabs_notification_channel),
+                            NotificationManager.IMPORTANCE_HIGH);
+                }
+                break;
+
+                case LP_DEFAULT: {
+                    channel = new NotificationChannel(mDefinedNotificationChannels.get(definedChannel),
+                            mContext.getString(R.string.leanplum_default_notifications_channel2),
+                            NotificationManager.IMPORTANCE_LOW);
+                }
+                break;
+
+                case SITE_NOTIFICATIONS: {
+                    channel = new NotificationChannel(mDefinedNotificationChannels.get(definedChannel),
+                            mContext.getString(R.string.site_notifications_channel),
+                            NotificationManager.IMPORTANCE_DEFAULT);
+                }
+                break;
+
+                case CRASH_HANDLER: {
+                    channel = new NotificationChannel(mDefinedNotificationChannels.get(definedChannel),
+                            mContext.getString(R.string.crash_handler_notifications_channel),
+                            AppConstants.Versions.feature29Plus ?
+                                    NotificationManager.IMPORTANCE_HIGH :
+                                    NotificationManager.IMPORTANCE_LOW);
+                }
+                break;
+
+                case DEFAULT:
+                default: {
+                    channel = new NotificationChannel(mDefinedNotificationChannels.get(definedChannel),
+                            mContext.getString(R.string.default_notification_channel2),
+                            NotificationManager.IMPORTANCE_LOW);
+                }
+                break;
+            }
+
+            manager.createNotificationChannel(channel);
+        }
+    }
+
+    @TargetApi(26)
+    private void removeChannel(final NotificationManager manager, final String channelId) {
+        if (manager.getNotificationChannel(channelId) != null) {
+            manager.deleteNotificationChannel(channelId);
+        }
+    }
+
+    @TargetApi(26)
+    public NotificationChannel getNotificationChannel(Channel definedChannel) {
+        final NotificationManager notificationManager = (NotificationManager) mContext.getSystemService(Context.NOTIFICATION_SERVICE);
+        return notificationManager.getNotificationChannel(mDefinedNotificationChannels.get(definedChannel));
+    }
+
     @Override // BundleEventListener
     public void handleMessage(final String event, final GeckoBundle message,
                               final EventCallback callback) {
@@ -182,6 +372,7 @@ public final class NotificationHelper implements BundleEventListener {
 
     private Intent buildNotificationIntent(final GeckoBundle message, final Uri.Builder builder) {
         final Intent notificationIntent = new Intent(HELPER_BROADCAST_ACTION);
+        notificationIntent.setClass(mContext, NotificationReceiver.class);
         final boolean ongoing = message.getBoolean(ONGOING_ATTR);
         notificationIntent.putExtra(ONGOING_ATTR, ongoing);
 
@@ -230,6 +421,7 @@ public final class NotificationHelper implements BundleEventListener {
         return res;
     }
 
+    @SuppressWarnings({"NewApi", "try"})
     private void showNotification(final GeckoBundle message) {
         ThreadUtils.assertOnUiThread();
 
@@ -267,6 +459,15 @@ public final class NotificationHelper implements BundleEventListener {
             builder.setLargeIcon(b);
         }
 
+        if (!AppConstants.Versions.preO) {
+            if ("downloads".equals(message.getString(HANDLER_ATTR))) {
+                builder.setChannelId(getNotificationChannel(Channel.DOWNLOAD).getId());
+                builder.setOnlyAlertOnce(true);
+            } else {
+                builder.setChannelId(getNotificationChannel(Channel.DEFAULT).getId());
+            }
+        }
+
         if (message.containsKey(PROGRESS_VALUE_ATTR) &&
             message.containsKey(PROGRESS_MAX_ATTR) &&
             message.containsKey(PROGRESS_INDETERMINATE_ATTR)) {
@@ -294,10 +495,15 @@ public final class NotificationHelper implements BundleEventListener {
         // scheme to prevent Fennec from popping up.
         final Intent viewFileIntent = createIntentIfDownloadCompleted(message);
         if (builder != null && viewFileIntent != null && mContext != null) {
-            final PendingIntent pIntent = PendingIntent.getActivity(
-                    mContext, 0, viewFileIntent, PendingIntent.FLAG_UPDATE_CURRENT);
-            builder.setAutoCancel(true);
-            builder.setContentIntent(pIntent);
+            // Bug 1450449 - Downloaded files already are already in a public directory and aren't
+            // really owned exclusively by Firefox, so there's no real benefit to using
+            // content:// URIs here.
+            try (StrictModeContext unused = StrictModeContext.allowAllVmPolicies()) {
+                final PendingIntent pIntent = PendingIntent.getActivity(
+                        mContext, 0, viewFileIntent, PendingIntent.FLAG_UPDATE_CURRENT);
+                builder.setAutoCancel(true);
+                builder.setContentIntent(pIntent);
+            }
 
         } else {
             final PendingIntent pi = buildNotificationPendingIntent(message, CLICK_EVENT);
@@ -315,6 +521,8 @@ public final class NotificationHelper implements BundleEventListener {
         // them when geckoapp is destroyed.
         if (!persistent && !mClearableNotifications.containsKey(id)) {
             mClearableNotifications.put(id, message);
+        } else if (persistent && mClearableNotifications.containsKey(id)) {
+            mClearableNotifications.remove(id);
         }
     }
 
